// Parses the provided fields for each parameter given in the paramList argument.
// It will parse by name if provided, otherwise it will parse by position, in the same order of paramList.
// Optional parameters contain the string " (optional)" after the parameter name, eg. "BYPASS (optional)".
// Returns false on error, otherwise returns an object mapping the parameter name to the value.
var parseParameters = function(paramList, fields) {
	var parameters = {};
	var optional = [];
	
	// Check if the parameter is optional
	for(var i=0; i<paramList.length; i++) {
		if(paramList[i].indexOf(" (optional)") !== -1) {
			paramList[i].replace(" (optional)", "");
			optional.push(true);
		} else {
			optional.push(false);
		}
	}
	
	var maxArgs = optional.length;
	var minArgs = optional.length - optional.filter(Boolean).length;	// optional.filter(Boolean).length returns number of true values in array
	
	if((fields.length-1) < minArgs || (fields.length-1) > maxArgs) {
		if(minArgs === maxArgs) {
			message = "Expected " + paramList.length + " arguments, found " + (fields.length - 1);
		} else {
			message = "Expected " + minArgs + " to " + maxArgs + " arguments, found " + (fields.length - 1);
		}
		throw(newError("ERR_WRONG_NUMBER_OF_ARGUMENTS", message, ""));
	}
    
	for(var i=0; i<paramList.length; i++) {
		// Parse the parameter
		if(optional[i] && fields[i+1] === undefined) {
			continue;
		} else {
			if(fields[i+1].indexOf("=") !== -1) {
				var parsedParameters = fields[i+1].split("=");
				parameters[parsedParameters[0].toUpperCase()] = parsedParameters[1];
			} else {
				parameters[paramList[i]] = fields[i+1];
			}
		}
	}
	
	for(i=0; i<paramList.length; i++) {
		if(!optional[i] && parameters[paramList[i]] === undefined) {
			var message = "ERROR: " + paramList[i] + " argument not found.  It must be argument #" + (i+1) + ", or else add '" + paramList[i] + "='";
			throw(newError("ERR_MISSING_ARGUMENT", message, ""));
		}
	}
	
	return parameters;
}

// Does error checking, and calculate address mapping from input address to PFSM address.
// Page 4 (WD) is converted to 0x3XX, and page 3 (SRAM) is unwritable and throws an error.
// Returns -1 on error.
var parseAddress = function(addressStr) {
	
    address = convertStringToInt(addressStr)
    
    if(isNaN(address)) {
		var message = "ERROR: Invalid register address, expecting a number";
		throw(newError("ERR_INVALID_ARGUMENT_VALUE", message, addressStr));
	}
	
	if(address >= 0x300 && address < 0x400) {
		var message = "ERROR: Cannot read or write to page 3 (SRAM) of address space.  Must be page 0, 1, 2, or 4";
		throw(newError("ERR_ADDRESS_PAGE_3", message, addressStr));
	}
    
    if(address > 0x4FF) {
		var message = "ERROR: Invalid register address, must be 10 bits or less. 11 bits is allowed only if writing to page 4";
		throw(newError("ERR_ADDRESS_OUT_OF_RANGE", message, addressStr));
	}
	
	// Maps WD page 4 in to PFSM page 3
	if(address >= 0x400) {
		address -= 0x100;
	}
	
	return address;
}

var parseNumber = function(valueStr, numBits, fieldName) {
     
    value = convertStringToInt(valueStr);
    
    if(isNaN(value)) {
		var message = "ERROR: Invalid " + fieldName + " value, expecting a number";
		throw(newError("ERR_INVALID_ARGUMENT_VALUE", message, valueStr));
    }
    
    if(value > (Math.pow(2, numBits) - 1)) {
		var message = "ERROR: Invalid " + fieldName + " value, must be " + numBits + " bits or less";
		throw(newError("ERR_INVALID_ARGUMENT_VALUE", message, valueStr));
    }
    
    return value;
}

var parseDelayStep = function(step, roundingTolerance) {
    
    var targetStep = convertTimeToNs(step);
    if(targetStep === -1) {
		var message = "ERROR: Invalid delay step '" + step + "'\nExpecting a number with units of ns, us, ms, or s";
		throw(newError("ERR_DELAY_BAD_FORMAT", message, step));
    }
    
    var delayFound = false;
    
    // Check if time step is too high or too low
    if(targetStep > 107.4e9) {
		var message = "ERROR: Invalid delay step '" + step + "'\nMaximum step size is 107.4 seconds";
		throw(newError("ERR_DELAY_OUT_OF_RANGE", message, step));
    } else if(targetStep < 50) {
		var message = "ERROR: Invalid delay step '" + step + "'\nMinimum step size is 50ns";
		throw(newError("ERR_DELAY_OUT_OF_RANGE", message, step));
    }
    
    var delayValue = 0;
    // Default value is 10%, to preserve backwards compatibility
    if(roundingTolerance === undefined) {
        roundingTolerance = 0.1;
    } else {
        roundingTolerance /= 100;
    }
    
    for(var j=0; j<32; j++) {
        var delayStep = 50 * Math.pow(2,j);
        // If specified delay is within 10% of a time step, round to that nearest step
        if(Math.abs(delayStep-targetStep) / delayStep < roundingTolerance) {
            delayValue = j;
            break;
        }
        // If we're larger than the expected time, it means we didn't find any within tolerance
        if(delayStep > targetStep) {
			var message = "ERROR: Invalid delay step: " + step + " with current rounding tolerance of " + (roundingTolerance*100) + "%.\nClosest values are " + convertNsToTime(50*Math.pow(2,j-1)) + " and " + convertNsToTime(50*Math.pow(2,j));
			throw(newError("ERR_DELAY_WRONG_STEP_SIZE", message, step));
        }
    }

    if(step !== convertNsToTime(delayStep)) {
        logMessage("INFO: specified delay step of " + step + " is converted to delay step of " + convertNsToTime(delayStep));
    }
    
    return delayValue;
}

var parseFieldFromMap = function(map, index, fieldName) {
    
    index = index.toUpperCase();
    if(!map.hasOwnProperty(index)) {
        optionsList = "";
        for(var property in map) {
            if(map.hasOwnProperty(property)) {
                optionsList += property + "\n";
            }
        }
		
		var message = "ERROR: Invalid " + fieldName + " '" + index + "'.  Options are:\n" + optionsList;
		throw(newError("ERR_INVALID_ARGUMENT_VALUE", message, index));
    }
    
    if(typeof(map[index]) === "string") {
        return parseInt(map[index], 2);
    } else {
        return map[index];
    }
}

var parseVoutToVset = function(voutStr, regulator, bypass, warnings) {

    var vout = 0;	// Vout value in mV
    
    // Validate VOUT is a number ending in mV or V
    if(voutStr.substring(voutStr.length-2).toUpperCase() === "MV") {
        vout = convertStringToInt(voutStr.substring(0, voutStr.length-2));
        if(isNaN(vout)) {
			var message = "ERROR: Invalid VOUT '" + voutStr + "', expecting a number with units of mV or V";
			throw(newError("ERR_INVALID_ARGUMENT_VALUE", message, voutStr));
        }
    } else if(voutStr.substring(voutStr.length-1).toUpperCase() === "V") {
        vout = convertStringToInt(voutStr.substring(0, voutStr.length-1))*1000;
        if(isNaN(vout)) {
			var message = "ERROR: Invalid VOUT '" + voutStr + "', expecting a number with units of mV or V";
			throw(newError("ERR_INVALID_ARGUMENT_VALUE", message, voutStr));
        }
    } else {
		var message = "ERROR: Invalid VOUT '" + voutStr + "', expecting a number with units of mV or V";
			throw(newError("ERR_INVALID_ARGUMENT_VALUE", message, voutStr));
    }
    
     var vset = 0;
    
    // Validate VOUT is an acceptable voltage value, and calculate the VSET value for that VOUT
    if(regulator.toUpperCase().indexOf("BUCK") !== -1) {
        if(vout < 300) {
			var message = "ERROR: Invalid VOUT '" + voutStr + "', must be between 0.3V and 3.34V for buck output voltage";
            throw(newError("ERR_INVALID_ARGUMENT_VALUE", message, voutStr));
        } else if(vout < 600) {
            vset = (vout-300) / 20;
            if(Math.round(vset) !== vset) {
                var message = "ERROR: Invalid VOUT '" + voutStr + "'. Nearest VSET values are " + (Math.floor(vset)*20 + 300) + "mV and " + (Math.ceil(vset)*20 + 300) + "mV";
				throw(newError("ERR_INVALID_ARGUMENT_VALUE", message, voutStr));
            }
        } else if(vout < 1100) {
            vset = ((vout-600) / 5) + 15;
            if(Math.round(vset) !== vset) {
                var message = "ERROR: Invalid VOUT '" + voutStr + "'. Nearest VSET values are " + ((Math.floor(vset)-15)*5 + 600) + "mV and " + ((Math.ceil(vset)-15)*5 + 600) + "mV";
				throw(newError("ERR_INVALID_ARGUMENT_VALUE", message, voutStr));
            }
        } else if(vout < 1660) {
            vset = ((vout-1100) / 10) + 0x73;
            if(Math.round(vset) !== vset) {
                var message = "ERROR: Invalid VOUT '" + voutStr + "'. Nearest VSET values are " + ((Math.floor(vset)-0x73)*10 + 1100) + "mV and " + ((Math.ceil(vset)-0x73)*10 + 1100) + "mV";
				throw(newError("ERR_INVALID_ARGUMENT_VALUE", message, voutStr));
            }
        } else if(vout <= 3340) {
            vset = ((vout-1660) / 20) + 0xAB;
            if(Math.round(vset) !== vset) {
                var message = "ERROR: Invalid VOUT '" + voutStr + "'. Nearest VSET values are " + ((Math.floor(vset)-0xAB)*20 + 1660) + "mV and " + ((Math.ceil(vset)-0xAB)*20 + 1660) + "mV";
				throw(newError("ERR_INVALID_ARGUMENT_VALUE", message, voutStr));
            }
        } else {
            var message = "ERROR: Invalid VOUT '" + voutStr + "', must be between 0.3V and 3.34V";
			throw(newError("ERR_INVALID_ARGUMENT_VALUE", message, voutStr));
        }
        
        if(bypass !== undefined) {
			var message = "WARNING: Bypass mode is not valid for BUCK regulators.  Ignoring bypass bit";
			var warn = newWarning("WARN_BUCK_BYPASS_IGNORED", message, "");
			warn.instruction = currentInstruction;
            warnings.push(warn);
        }
        
    } else { // LDO
        if (vout < 600) {
			var message = "ERROR: Invalid VOUT '" + voutStr + "', must be between 0.6V and 3.3V for LDO output voltage";
			throw(newError("ERR_INVALID_ARGUMENT_VALUE", message, voutStr));
        } else if(vout <= 3300) {
            var ldo = regulator.toUpperCase();
            if(ldo === "LDO4") {
                vset = ((vout-600) / 25) + 0x8;
                if(Math.round(vset) !== vset) {
                    var message = "ERROR: Invalid VOUT '" + voutStr + "'. Nearest VSET values are " + ((Math.floor(vset)-0x8)*25 + 600) + "mV and " + ((Math.ceil(vset)-0x8)*25 + 600) + "mV";
					throw(newError("ERR_INVALID_ARGUMENT_VALUE", message, voutStr));
                }	
            } else { // LDO1-3
                vset = ((vout-600) / 50) + 0x4;
                if(Math.round(vset) !== vset) {
                    var message = "ERROR: Invalid VOUT '" + voutStr + "'. Nearest VSET values are " + ((Math.floor(vset)-0x4)*50 + 600) + "mV and " + ((Math.ceil(vset)-0x4)*50 + 600) + "mV";
					throw(newError("ERR_INVALID_ARGUMENT_VALUE", message, voutStr));
                }					
                vset = vset << 1;
            }

        } else {
            var message = "ERROR: Invalid VOUT '" + voutStr + "', must be between 0.6V and 3.3V for LDO output voltage";
			throw(newError("ERR_INVALID_ARGUMENT_VALUE", message, voutStr));
        }
        
        if(bypass !== undefined && bypass !== "0" && bypass !== "1") {
            var message = "ERROR: Bypass mode must be 1 or 0 when writing an LDO output voltage";
			throw(newError("ERR_INVALID_ARGUMENT_VALUE", message, voutStr));
        } else {
            if(bypass === "1") {
                vset |= 0x80;
            }
        }
    }
    
    return vset;
}

// Converts a delay time in to a delay value, given a certain delay step.  It will round to the nearest delay value.
// numBits is the number of bits for the delay field.  Some instructions have 6 bits, some have 8 bits.
// For example, delayTime=5ms, delayStep=1.6ms, it will return 3.
// Returns -1 on error
var parseDelay = function(delayTime, delayStep, numBits, roundingTolerance, warnings) {
	
	// If it's a number already, just return the number
	if(!isNaN(delayTime))  {
		return convertStringToInt(delayTime);
	}
	targetDelayTimeNs = convertTimeToNs(delayTime);
	if(delayStep === -1 && targetDelayTimeNs !== 0) {
		var message = "ERROR: No delay step is set.  Define default PFSM_DELAY_STEP value, or use SET_DELAY command before this instruction";
		throw(newError("ERR_DELAY_STEP_NOT_SET", message, ""));
	}
	if(targetDelayTimeNs === -1) {
		var message = "ERROR: Invalid delay time '" + delayTime + "', expecting a number with units of ns, us, ms, or s";
		throw(newError("ERR_DELAY_BAD_FORMAT", message, delayTime));
	}
	if(targetDelayTimeNs < 0) {
		var message = "ERROR: Invalid delay time '" + delayTime + "', must not have negative delay";
		throw(newError("ERR_DELAY_OUT_OF_RANGE", message, delayTime));
	}
	if(targetDelayTimeNs > (Math.pow(2,numBits)-1) * 107.4e9) {
		var message = "ERROR: Invalid delay time '" + delayTime + "', maximum delay is " + ((Math.pow(2, numBits)-1) * 107.4) + " seconds";
		throw(newError("ERR_DELAY_OUT_OF_RANGE", message, delayTime));
	}
    
    var delayVal = Math.round(targetDelayTimeNs / delayStep);
    var actualDelayTimeNs = delayStep * delayVal;
    roundingTolerance /= 100;
	if(delayVal === 0 && targetDelayTimeNs !== 0) {
		var message = "WARNING: Delay time of " + delayTime + " is rounded to 0, with current delay step of " + convertNsToTime(delayStep);
		var warning = newWarning("WARN_DELAY_ROUND_TO_ZERO", message, "");
		warning.instruction = currentInstruction;
		warnings.push(warning);
	} else if(delayVal > Math.pow(2, numBits)-1) {
		var message = "ERROR: Delay time of " + delayTime + " cannot be achieved with current step size of " + convertNsToTime(delayStep) + "\nUse SET_DELAY command to choose a larger delay step size.";
		throw(newError("ERR_DELAY_WRONG_STEP_SIZE", message, delayTime));
	} else if(roundingTolerance !== undefined && Math.abs((targetDelayTimeNs-actualDelayTimeNs) / actualDelayTimeNs) > roundingTolerance) {
		var message = "WARNING: Delay time of " + delayTime + " cannot be achieved with rounding tolerance of " + (roundingTolerance*100) + "%.\nClosest value is " + convertNsToTime(actualDelayTimeNs);
		var warning = newWarning("WARN_DELAY_NOT_WITHIN_TOLERANCE", message, "");
		warning.instruction = currentInstruction;
		warnings.push(warning);
    }
	
	return delayVal;
}

// Converts number of nanoseconds in to a time including units
var convertNsToTime = function(time) {
	if(time < 1e3) {
		return time + "ns";
	}
	if(time < 1e6) {
		return +((time/1e3).toFixed(3)) + "us"; // This rounds to 3 digits, but drops trailing 0s
	}
	if(time < 1e9) {
		return +((time/1e6).toFixed(3)) + "ms";
	}
	return +((time/1e9).toFixed(3)) + "s";
}

// Converts the given time (including unit) in to nanoseconds
var convertTimeToNs = function(timeStr) {
	if(timeStr.substring(timeStr.length-2).toLowerCase() == "ns") {
		time = convertStringToInt(timeStr.substring(0, timeStr.length-2));
		if(isNaN(time)) {
			return -1;
		} else {
			return time;
		}
	}
	
	if(timeStr.substring(timeStr.length-2).toLowerCase() == "us") {
		time = convertStringToInt(timeStr.substring(0, timeStr.length-2));
		if(isNaN(time)) {
			return -1;
		} else {
			return time * 1e3;
		}
	}
	
	if(timeStr.substring(timeStr.length-2).toLowerCase() == "ms") {
		time = convertStringToInt(timeStr.substring(0, timeStr.length-2));
		if(isNaN(time)) {
			return -1;
		} else {
			return time * 1e6;
		}
	}
	
	if(timeStr.substring(timeStr.length-1).toLowerCase() == "s") {
		time = convertStringToInt(timeStr.substring(0, timeStr.length-1));
		if(isNaN(time)) {
			return -1;
		} else {
			return time * 1e9;
		}
	}
	
	return -1;
}

var parseDestinationToAddress = function(destination, symbolsAddress) {
    
    if(!symbolsAddress.hasOwnProperty(destination)) {
		var message = "ERROR: Invalid destination '" + destination + "'";
        throw(newError("ERR_INVALID_DESTINATION_NAME", message, destination));
    }
    
    if (symbolsAddress[destination] > 0x1FFF) {
		var message = "ERROR: Destination address 0x" + symbolsAddress[destination].toString(16) + " exceeds 13 bits";
		throw(newError("ERR_INVALID_DESTINATION_ADDRESS", message, destination));
    }
    
    return symbolsAddress[destination];
}

var parseDestinationToSkipCount = function(destination, symbolsInstructionCount, currentInstructionCount) {
    
    if(!symbolsInstructionCount.hasOwnProperty(destination)) {
		var message = "ERROR: Invalid destination '" + destination + "'";
		throw(newError("ERR_INVALID_DESTINATION_NAME", message, destination));
    }
    
    var skipCount = symbolsInstructionCount[destination] - currentInstructionCount - 1;
    if(skipCount < 0 || skipCount > 63) {
		var message = "ERROR: Invalid destination '" + destination + "'.\n  Destination is " + skipCount + " instructions after the WAIT statement, must be between 0 and 63.";
		throw(newError("ERR_INVALID_DESTINATION_ADDRESS", message, destination));
    }
    
    return skipCount;
}

var parseBit = function(bitStr, fieldName) {
    
    bitVal = convertStringToInt(bitStr);
    if(isNaN(bitVal) || bitVal < 0 || bitVal > 1) {
		var message = "ERROR: Invalid " + fieldName + " value '" + bitStr + "', must be 0 or 1";
		throw(newError("ERR_INVALID_ARGUMENT_VALUE", message, bitStr));
    }
    
    return bitVal;
}

var parseTriggerId = function(trigId) {
    
    trigIdVal = convertStringToInt(trigId);
    if(isNaN(trigIdVal) || trigIdVal > 27 || trigIdVal < 0) {
		var message = "ERROR: Invalid trigger ID '" + trigId + "', expecting a number between 0 and 27";
        throw(newError("ERR_INVALID_ARGUMENT_VALUE", message, trigId));
    }
    
    return trigIdVal;
}

var parseLiteralCommand = function(fields) {
	var binary = "";
	
	for(j=1; j<fields.length; j++) {
		
		var argument = "";
		
		if(fields[j].indexOf("=") === -1) {
			argument = fields[j];
		} else {
			argument = fields[j].split("=")[1];
		}
		
		if(argument.indexOf(",") === -1) {
			var message = "ERROR: Field '" + fields[j] + "', is missing the number of bits.  The format should be '<#bits>,<value>'";
			throw(newError("ERR_INVALID_LITERAL", message, ""));
		}
		
		var argValue = argument.split(",");
		var numBits = convertStringToInt(argValue[0]);
		var value = convertStringToInt(argValue[1]);
		
		if(isNaN(numBits)) {
			var message = "ERROR: Invalid number of bits: " + argValue[0];
			throw(newError("ERR_INVALID_LITERAL", message, ""));
		}
		
		if(isNaN(value)) {
			var message = "ERROR: Invalid value: " + argValue[1];	
			throw(newError("ERR_INVALID_LITERAL", message, ""));
		}
		
		if(value.toString(2).length > numBits) {
			var message = "ERROR: Invalid value: " + argValue[1] + ", it is longer than the specified bit size";
			throw(newError("ERR_INVALID_LITERAL", message, ""));
		}
		
		binary = zeroPad(value.toString(2), numBits) + binary;
	}
	
	return binary;
}

// Our jjs engine (to run from command line) doesn't support Number() conversion leading with 0b.
// So this will use parseInt if a binary value is proveded.
var convertStringToInt = function(str) {
	if(str.substring(0, 2) == "0b") {
		return parseInt(str.substring(2), 2);
	}
	else {
		return Number(str);
	}
}